import * as React from 'react';
import * as ReactDOM from 'react-dom';

// Чтобы не материлось Гы Гы Гы
declare function require(moduleName: string): any;

import { ui } from './ui';
const classesMaps: any = require('ts/classes/classesMaps.json');

/**Параметры функции по созданию объекта "из файла" */
export interface CreateObjectFromFileParams {
  /**Ключ вида из коллекции определений (тех, что описаны в файле json) */
  keyView?: string;
  /**Дополнительные параметры контроллера */
  controllerParams?: Object;
}

/**Параметры функции по созданию вида "из файла" */
export interface CreateElementFromFileParams {
  /**Ключ для добавления вида в коллекцию React */
  key?: any;
  /**Дополнительные параметры вида */
  viewerParams?: Object;
}

export class CommonUtils {
  /**Из пути к классу возвращает имя класса - 
   * если имя dir1/dir2/fileName - вернет fileName;
   * если имя dir1/dir2/fileName$className - вернет className
  */
  public static getClassName(pathToClass: string): string {
    // Проверю на наличие спец символа $
    let lastIndex = pathToClass.lastIndexOf('$');
    if (lastIndex >= 0) {
      return pathToClass.substr(lastIndex + 1);
    }
    lastIndex = pathToClass.lastIndexOf('/');
    if (lastIndex >= 0) {
      pathToClass = pathToClass.substr(lastIndex + 1);
    }
    lastIndex = pathToClass.lastIndexOf('.');
    if (lastIndex >= 0) {
      pathToClass = pathToClass.substr(lastIndex + 1);
    }
    return pathToClass;
  };

  /**Из пути к классу возвращает путь к файлу - 
   * если имя dir1/dir2/fileName - вернет dir1/dir2/fileName;
   * если имя dir1/dir2/fileName$className - вернет dir1/dir2/fileName
   */
  public static getFileName(pathToClass: string): string {
    // Проверю на наличие спец символа $
    let lastIndex = pathToClass.lastIndexOf('$');
    if (lastIndex >= 0) {
      return pathToClass.substr(0, lastIndex);
    }
    return pathToClass;
  };

  /** Создает объект заданного класса по описанию из файла
   * Если в файле хранится объявление класса - будет создан объект описанного класса
   * Если в файле содержиться определение класса вида export default {title: "Бла Бла Бла"} 
   * будет создан объект заданного класса с передачей в его конструктор определения класса
   * 
   * cls - Функция конструктора класса (т.е. имя класса в понимании typescript)
   * ownerController - контроллер владелец создаваемого класса
   * pathToClass - путь к классу определения файла начиная с ts (может иметь вид ui/ui, а может ui/ui$Class)
   * 
   * Пример вызова let obj = CommonUtils.createObjectFromFile(Class, module, 'ui/classes/Transport/TransportList');
  */
  public static createObjectFromFile<T>(
    /* Функция конструктор класса контроллера */
    cls: { new (ownerController: any, classDef?: Object, params?: Object): T; },
    /** Родительский контроллер */
    ownerController: any,
    /**Путь к определению класса */
    pathToClass: string,
    params?: CreateObjectFromFileParams
  ): T {

    console.warn('SIMA CommonUtils.createObjectFromFile ', pathToClass, params);

    let keyView: string = '';
    if (params && params.keyView) {
      keyView = params.keyView;
    }

    let controllerParams: Object = {};
    if (params && params.controllerParams) {
      controllerParams = params.controllerParams;
    }

    let className = CommonUtils.getClassName(pathToClass);
    let fileName = CommonUtils.getFileName(pathToClass);

    // console.log('SIMA createObjectFromFile', pathToClass);
    let classJson: any = CommonUtils.getClassJson(fileName, className, keyView);
    let classDef: any = null;

    // Поиск класса определения
    let fn: string = '';
    if (classJson.proxyClass) {
      // Для случая, когда для заданного класса в classesMaps.json задан заместитель (className)
      fileName = CommonUtils.getFileName(classJson.proxyClass);
      className = CommonUtils.getClassName(classJson.proxyClass);
      fn = './' + fileName + '.ts';
    }
    else {
      fn = './classes/ts/' + fileName + '.ts';
    }

    try {
      classDef = require(fn);
    }
    catch (expn) {
      console.log('SIMA !!!! Ошибка создания. Модуль ' + pathToClass + ' найти не удалось');
      // console.warn('CommonUtils.createObjectFromFile', expn, fn);
    }

    if (classDef === null) {
      // Если classDef не определен - воспользуемся базовым классом
      cls = ui.Class as any;
    }
    else if (classDef && classDef.hasOwnProperty('__esModule') && classDef.__esModule) {
      // Прикол в том, что в загруженном модуле может быть определен namespace
      // в это случае имена классов должны и будут составными к примеру такими ui.Class
      // поэтому, вначале необходимо "погрузиться" во все namespace'ы, и уж потом создавать класс
      cls = CommonUtils.getCtor(classDef, className);
    }
    if (cls) {
      return new cls(ownerController, classJson, controllerParams);
    }
    throw new Error('SIMA Не удалос создать объект для' + pathToClass);
  }

  private static getClassJson(fileName: string, className: string, keyView: string): any {
    // Поиск файла структуры класса
    // Мар на файл
    let map: any;
    // map = classesMaps[fileName];
    // Ключ класса
    let classKey: string = fileName + '$' + className;
    // Пытаюсь найти по полному имени
    map = classesMaps[classKey];
    if (!map) {
      // если не удалось ищу по имени файла, считая что класс будет одноименным с именем файла
      classKey = fileName;
      map = classesMaps[classKey];
    }
    console.warn('SIMA ================ MAP ', classKey, className, (map ? '; map из classesMaps.json найден' : '; !!!!! map из classesMaps.json НЕ найден'));

    let classJson: any = null;
    // Первым делом в папке classes/jsJson
    let path: string = './classes/jsJson/' + classKey + '.json';
    try {
      classJson = require(path);
      classJson = CommonUtils.convertJavaClassesToJS(classJson);
    }
    catch (expn) {
      // console.log('SIMA !!!! 1 CommonUtils.getClassJson. Не удалось найти модуль ' + path);
    }
    // Если не найдено, то в папке classes/javaJson
    if (classJson === null) {
      path = './classes/javaJson/' + classKey + '.json';
      try {
        classJson = require(path);
        // если найдено, проведем конвертацию из "java" формата в "js" формат
        classJson = CommonUtils.convertJavaClassesToJS(classJson);
      }
      catch (expn) {
        console.log('SIMA !!!! 2 CommonUtils.getClassJson. Не удалось найти модуль ' + path);
        // console.warn(expn);
      }
    }

    /// Выберем заданное определение по keyView
    if (classJson && Array.isArray(classJson)) {
      const arr = classJson;
      for (let key in arr) {
        let item = arr[key];
        let clsKeyView: string = item.keyView || '';
        if (clsKeyView === keyView) {
          classJson = item;
          break;
        }
      }
    }

    // Добавим свойства из MAP'а
    if (map) {
      if (!classJson) {
        classJson = {};
      }
      for (let key in map) {
        if (!classJson.hasOwnProperty(key)) {
          classJson[key] = map[key];
        }
      }
    }

    return classJson;
  }

  /**Преобразует java описание в формат JS 
   * На вход получу массив описаний
  */
  public static convertJavaClassesToJS(j: Array<any>): Array<any> {
    const result = new Array<any>();
    if (!Array.isArray(j)) {
      return j;
    }
    for (let idx in (j as Array<any>)) {
      result.push(CommonUtils.convertJavaClassToJS(j[idx]));
    }
    return result;
  }

  private static convertJavaClassToJS(j: any): any {
    const result: any = {};
    for (let key in j) {
      switch (key) {
        case 'title':
        case 'keyView':
          result[key] = j[key];
          break;
        case 'fields':
          result[key] = CommonUtils.convertJavaFieldsToJS(j[key]);
          break;
        case 'actions':
          result[key] = CommonUtils.convertJavaClassActionsToJS(j[key]);
          break;
      }
    }
    return result;
  }

  private static convertJavaFieldsToJS(j: any): any {
    const result = new Array<any>();
    if (!Array.isArray(j)) {
      return j;
    }
    for (let idx in (j as Array<any>)) {
      result.push(CommonUtils.convertJavaFieldToJS(j[idx]));
    }
    return result;
  }

  private static convertJavaFieldToJS(j: any) {
    const result: any = {};
    for (let key in j) {
      switch (key) {
        case 'beanPropertyName':
        case 'title':
        case 'order':
        case 'content':
        case 'readOnly':
        case 'showInManagerPanel':
        case 'visible':
        case 'propertiesLayout':
          result[key] = j[key];
          break;
        case 'typeField':
          /// Здесь, скорее всего надо делать по-другому, а имеено, через заполнение 2-х полей класса wsType и jsonType
          result[key] = CommonUtils.convertJavaTypeToJSType(j[key]);
          break;
        case 'viewer':
          result[key] = CommonUtils.findViewer(j[key]);
          break;
        case 'panels':
          result[key] = CommonUtils.convertJavaPanelsToJS(j[key]);
          break;
        case 'propertiesField':
          result['propertiesField'] = CommonUtils.convertPropertiesField(j[key]);
          // console.log('SIMA vvvvvvvvvvvvvvvvvvvvvvvv', result['propertiesField']);
          break;
      }
    }
    return result;
  }

  /** Конвертирует строковый массив в объект с очиской значений от мусорных символов 
   * 
  */
  private static convertPropertiesField(j: any): any {
    const retVal: any = {};
    if (!Array.isArray(j)) {
      return retVal;
    }
    const a = j as Array<string>;
    for (let idx in a) {
      let str = a[idx].replace('\\u003d', '=');
      // Вот такой вот черезжопный метод очистки от лишних знаков \t и \n
      // через Regexp не получилось..., а стандартный replace удаляет почему-то только одно найденное вхождение ((
      // уже позже узнал что в javascript'овом regexp есть волшебный флаг типа g - globals
      // поэтому скорее всего шаблон /\\t/g мог бы и помочь, но переделывать уже работающий код не буду... 
      let str1 = str.replace('\\t', '');
      while (str.length !== str1.length) {
        str = str1;
        str1 = str.replace('\\t', '');
      }
      str1 = str.replace('\\n', '');
      while (str.length !== str1.length) {
        str = str1;
        str1 = str.replace('\\n', '');
      }
      let parts: string[] = str.split('=');
      if (parts[0] === 'clazz') {
        retVal['viewer'] = this.findViewer(parts[1]);
      }
      else if (parts[0] === 'contentclazz') {
        retVal['content'] = parts[1];
      }
      else {
        retVal[parts[0]] = parts[1];
      }
    }
    return retVal;
  }

  private static convertJavaTypeToJSType(j: string): string {
    if (j.startsWith('[L')) {
      j = j.replace('[L', '');
    }
    j = j.replace(';', '');
    switch (j) {
      case 'java.lang.String':
        return 'string';

      case 'java.util.List':
        return 'array';       // !!!!!!!!!!!!!!!!

      case 'ru.infor.beanviewer.viewer.BeanViewerAction':
        return 'action';      // !!!!!!!!!!!!!!!!

      case 'java.util.Date':
        return 'date';        // !!!!!!!!!!!!!!!!

      case 'java.lang.Integer':
      case 'int':
        return 'number';

      case 'boolean':
      case 'java.lang.Boolean':
        return 'boolean';

      case 'java.lang.Long':
        return 'number';

      case 'ru.blok.vms.ui.mapviewer.mapviewerwithtrack.MapValueWithTrack':
        return 'XXXXXXXXXX';  // !!!!!!!!!!!!!!!!

      case 'java.lang.Object':
        return 'object';

      case 'java.lang.Double':
      case 'double':
        return 'number';

      case 'java.math.BigDecimal':
        return 'number';

      case '':
        return '';

      case '':
        return '';

      default:
        return j;
    }
  }

  /**Подменяет указанное в json имя java класса вида на имя класса вида текущего приложения  */
  private static findViewer(j: string): string {
    if (Array.isArray(j)) {
      j = j[0];
    }
    let map: any = classesMaps[j];
    if (map && map.proxyClass) {
      return map.proxyClass;
    }
    return j;
  }

  private static convertJavaPanelsToJS(j: any): any {
    const result = new Array<any>();
    if (!Array.isArray(j)) {
      return j;
    }
    for (let idx in (j as Array<any>)) {
      result.push(CommonUtils.convertJavaPanelToJS(j[idx]));
    }
    return result;
  }

  private static convertJavaPanelToJS(j: any) {
    const result: any = {};
    for (let key in j) {
      switch (key) {
        case 'name':
        case 'order':
          result[key] = j[key];
          break;
        case 'classView':
          if (Array.isArray(j[key]) && (j[key] as Array<any>).length > 0) {
            result['controller'] = j[key][0].clazz;
          }
          break;
        case 'actions':
          result[key] = CommonUtils.convertJavaPanelActionsToJS(j[key]);
          break;
      }
    }
    return result;
  }

  private static convertJavaPanelActionsToJS(j: any): any {
    const result = new Array<any>();
    if (!Array.isArray(j)) {
      return j;
    }
    for (let idx in (j as Array<any>)) {
      result.push(CommonUtils.convertJavaPanelActionToJS(j[idx]));
    }
    return result;
  }

  private static convertJavaPanelActionToJS(j: any) {
    const result: any = {};
    for (let key in j) {
      switch (key) {
        case 'keyView':
        case 'order':
        case 'doubleClick':
        case 'clazz':
          result[key] = j[key];
          break;
        // case 'actionsUI':
        //   result['controller'] = j[key];
        //   break;
      }
    }
    return result;
  }

  private static convertJavaClassActionsToJS(j: any): any {
    const result = new Array<any>();
    if (!Array.isArray(j)) {
      return j;
    }
    for (let idx in (j as Array<any>)) {
      result.push(CommonUtils.convertJavaClassActionToJS(j[idx]));
    }
    return result;
  }

  private static convertJavaClassActionToJS(j: any) {
    const result: any = {};
    for (let key in j) {
      switch (key) {
        case 'name':
        case 'order':
        case 'icon':
        case 'showInManagerPanel':
        case 'showInPanel':
        case 'hint':
        case 'clazz':
          result[key] = j[key];
          break;
        // case 'clazz':
        //   result['controller'] = j[key];
        //   break;
      }
    }
    // console.log('SIMA )))))))))))))))))))))))))))))))))))))))))))))))))))', j, result);
    return result;
  }

  /**Создает объект действие */
  public static createAction(
    /**Контроллер владелец */
    ownerController: ui.IPanel | ui.IClass,
    /**Определение действия */
    acnDef: any,
    /**Параметры создаваемого контроллера действия */
    controllerParams?: Object): ui.IAction {

    let clazz: string = acnDef.clazz;
    let fileName: string = null;
    let className: string = null;
    let classDef: any = null;
    let classJson: any = null;
    if (clazz) {
      fileName = CommonUtils.getFileName(clazz);
      className = CommonUtils.getClassName(clazz);
      classJson = CommonUtils.getClassJson(fileName, className, (acnDef.keyView || ''));
    }
    // Здесь "объединим" определения из разных источников
    let resultJson: any = {};
    for (let key in acnDef) {
      resultJson[key] = acnDef[key];
    }
    if (classJson) {
      for (let key in classJson) {
        resultJson[key] = classJson[key];
      }
    }

    // console.log('SIMA createAction resultJson', fileName, className, resultJson);

    // Поиск класса определения
    let fn: string = '';
    if (resultJson.proxyClass) {
      // Для случая, когда для заданного класса в classesMaps.json задан заместитель (proxyClass)
      fileName = CommonUtils.getFileName(resultJson.proxyClass);
      className = CommonUtils.getClassName(resultJson.proxyClass);
      fn = './' + fileName + '.ts';
    }
    else {
      fn = './classes/ts/' + fileName + '.ts';
    }
    try {
      classDef = require(fn);
    }
    catch (expn) {
      console.warn(expn, fn);
    }
    let cls: { new (ownerController: any, classDef?: Object, params?: Object): ui.IAction };

    if (classDef === null) {
      cls = ui.Action;
    }
    else if (classDef && classDef.hasOwnProperty('__esModule') && classDef.__esModule) {
      // Прикол в том, что в загруженном модуле может быть определен namespace
      // в это случае имена классов должны и будут составными к примеру такими ui.Action
      // поэтому, вначале необходимо "погрузиться" во все namespace'ы, и уж потом создавать класс
      cls = CommonUtils.getCtor(classDef, className);
    }
    if (cls) {
      return new cls(ownerController, resultJson, controllerParams);
    }
    throw new Error('SIMA CommonUtils.createAction. Не удалось создать объект для' + acnDef);

  }

  /** Создает компонент заданного класса  */
  public static createElementFromFile<T extends ui.IController>(
    /**Вьювер, "для которого" создается элемент */
    viewer: ui.IViewer<T>,
    /**Функция обратного вызова, которая будут вызвана на создание реального компонента React */
    ref_onComponentChanged: (component: any) => void,
    /**Файл, из которого необходимо создать компонент */
    filePath: string
  ): JSX.Element {

    console.warn('SIMA createElementFromFile', filePath, viewer);

    let viewerName = CommonUtils.getClassName(filePath);
    let fileName = CommonUtils.getFileName(filePath);

    let viewerDef: any = null;
    try {
      viewerDef = require('ts/' + fileName + '.tsx');
    }
    catch (expn) {
      viewerDef = null;
    }
    if (viewerDef === null) {
      throw new Error('Компонент ' + fileName + ' не найден');
      // return null;
    }

    if (viewerDef.hasOwnProperty('__esModule') && viewerDef.__esModule) {
      if (viewerDef[viewerName] && typeof viewerDef[viewerName] === 'function') {

        let props: any = {
          viewer: viewer,
          ref: ref_onComponentChanged
        };
        // Обработка специального ключа передаваемого в params - key, 
        // который используется React'ом для задания ключа элеметнов массива или перечислений
        if (viewer.viewerParams && viewer.viewerParams.key) {
          // Передадим в создаваемый элемент ключ (key)
          props['key'] = viewer.viewerParams.key;
        }
        if (1 !== 1) {
          console.log('SIMA createElementFromFile --- ', viewerDef[viewerName], props);
          return React.createElement('div', null, 'getElement2');
        }
        let obj = React.createElement(viewerDef[viewerName], props);
        return obj;
      }
    }
    throw new Error('SIMA CommonUtils.createViewerFromFile. Не удалось создать объект для' + filePath);
  }


  /**Из модуля извлекает функцию конструктор для класса по пути к классу */
  private static getCtor(module: any, className: string): any {
    let m: any = module;
    let cn: string = className;
    let mn: string = '';
    let idx: number = cn.indexOf('.');
    while (idx >= 0) {
      mn = cn.substr(0, idx);
      if (m[mn]) {
        m = m[mn];
        cn = cn.substr(idx + 1);
        idx = cn.indexOf('.');
      }
      else {
        return null;
      }
    }
    return m[cn];
  }
}