import * as React from 'react';
import * as ReactDOM from 'react-dom';

import { ui } from '../ui';
import { event } from '../core/Event';
// import {AutoSizer} from 'components'; // Такое подключение вызывает матершину "компилятора"
// такое - нет
declare function require(moduleName: string): any;
let AutoSizer = require('components').AutoSizer;
let css = require('./OverlapStackPanel.scss');

export namespace osp {

  // ==============================================================================
  // =================================== События ==================================
  // ==============================================================================

  // /**Класс реализующий аргументы события ComponentChanged */
  // export class OverlapStackPanelsChangedEventArgs extends event.EventArgs {
  //   constructor(component: any) {
  //     super();
  //     this._component = component;
  //   }

  //   private _component: any;
  //   get component(): any {
  //     return this._component;
  //   }
  // }

  // export class OverlapStackPanelsChangedEvent extends event.Event<OverlapStackPanelsChangedEventArgs> { }

  // =========================== OverlapStackPanelProps =============================
  /**"Публичные" свойства компонента OverlapStackPanel */
  export interface IOverlapStackPanelsProps {
    /**Callback, через который компонент возвращает ссылку на объект управления  */
    initialized?: (controller: IOverlapStackPanels) => void;
  }

  // ============================= IOverlapStackPanels ==============================
  /**Интерфейс контроллера стека перекрываемых панелей */
  export interface IOverlapStackPanels {

    /**Индексатор */
    getPanelController(idx: number): IOverlapPanelController;

    /**Возвращает количество панелей в стеке (только чтение) */
    panelsCount: number;

    /**Добавляет компонент в стек перекрываемых панелей и возвращает строковый ключ панели */
    addPanel(element: JSX.Element, controller?: IOverlapPanelController): string;

    /**Удаляет последнюю панель из стека и возвращает строковый ключ удаленной панели*/
    removePanel(): string;

    /**Событие на изменение содержимого стека перекрываемых панелей */
    overlapStackPanelsChanged: event.Event<event.EventArgs>;
  }

  // =========================== IOverlapPanelController ============================
  /**Интерфейс контроллера компонента панели 
   * Реализуется классами, которые желают управлять открытием новых и закрытием существующих 
   * панелей
  */
  export interface IOverlapPanelController {
    /**Ссылка на контроллер стека перекрываемых панелей */
    overlapStackPanels: IOverlapStackPanels;

    /**Дружественное имя панели */
    title: string;

    /**Метод, возвращающий признак возможности закрытия (удаления) панели из стека */
    canClose(panelKey?: string): boolean;

    /**Возвращает элемент управления для панели */
    getDashboard(): JSX.Element;
  }


  // ============================== IOverlapPanelInfo ===============================
  /**Информационный класс, объект которого хранит информацию о добавленной панели */
  interface IOverlapPanelInfo {
    /**Ключ панели */
    panelKey: string;

    /**RJX элемент, который был подан на создание панели */
    element: JSX.Element;

    /**Ссылка на объект компонента, если RJX элемент представляет собой React.Component, т.е. на JS класс */
    controller: IOverlapPanelController;
  }


  // ============================= OverlapPanelInfo =================================
  /**Информационный класс, объект которого хранит информацию о добавленной панели */
  class OverlapPanelInfo implements IOverlapPanelInfo {

    constructor(panelKey: string, element: JSX.Element, controller?: IOverlapPanelController) {
      this._panelKey = panelKey;
      this.element = element;
      if (!controller) {
        this.controller = null;
      }
      else {
        this.controller = controller;
      }
    }

    _panelKey: string;
    /**Ключ панели */
    get panelKey(): string {
      return this._panelKey;
    }

    _element: JSX.Element;
    /**JSX элемент */
    get element(): JSX.Element {
      return this._element;
    }
    set element(value: JSX.Element) {
      this._element = value;
    }

    _controller: IOverlapPanelController;
    /**Соответствующий элементу контроллер*/
    get controller(): IOverlapPanelController {
      return this._controller;
    }
    set controller(value: IOverlapPanelController) {
      this._controller = value;
    }
  }

  // =================================== State ======================================
  /**Класс задающий структуру объекта состояния компонета */
  class State {
    constructor() {
      this.panels = [];
      this.divId = OverlapStackPanels.generateId();
    }
    divId: string;
    panels: IOverlapPanelInfo[];
  }

  // ============================= OverlapStackPanel ================================
  /**Компонент перекрываемого стека панелей */
  export class OverlapStackPanels extends React.Component<IOverlapStackPanelsProps, State> implements IOverlapStackPanels {

    constructor(props: IOverlapStackPanelsProps) {
      super(props);
      this._overlapStackPanelsChanged = new event.Event<event.EventArgs>();
      this.state = new State();
    }

    /**Возвращает компонент по его индексу */
    getPanelController(idx: number): IOverlapPanelController {
      if (idx < 0 || idx >= this.panelsCount) {
        throw new Error('Индекс выходит за допустимые рамки');
      }
      return this.state.panels[idx].controller;
    }

    get panelsCount(): number {
      return this.state && this.state.panels ? this.state.panels.length : 0;
    }

    /**Добавляет панель с клоном переданного элемента  и делает панель верхней, Возвращает panelKey */
    addPanel(element: JSX.Element, controller?: IOverlapPanelController): string {
      let panelKey: string = (element && element.props && element.props.panelKey ? element.props.panelKey : OverlapStackPanels.generateId());
      let panel: IOverlapPanelInfo;
      if (element && typeof element.type === 'function') {
        console.log('SIMA addPanel 1', element);
        /*
        // ТАКОЙ ФИНТ не прошел ((((( все заморожено !!!!!!!!!
        // Что примечательно, так это то, что e.props.controller - уже не заморожен, но как его применить к нашей ситуации... ума не приложу
        let e: any = element;
        if( e.hasOwnProperty('props') && e.props != null && !Object.isFrozen(e.props)  ){
          e.props.ref = this.ref_registerChild;
        }
        */

        panel = new OverlapPanelInfo(panelKey, element, controller);

        /*  
        panel = new OverlapPanelInfo(panelKey, React.cloneElement(element, {
          // Передадим туда ключ панели
          panelKey: panelKey,
          // Передадим туда функцию обратного вызова для регистрации объекта js компонента (а не JSX)
          // ref - это стандартный React'овский механизм получать ссылку на объект компонента и вызывается самой инфраструктурой React после DidMount
          ref: this.ref_registerChild
        }));
        */
      }
      else {
        console.log('SIMA addPanel 2', element);
        panel = new OverlapPanelInfo(panelKey, element, controller);
      }

      if (1 === 1) return panelKey;

      this.state.panels.push(panel);
      this.setState({ panels: this.state.panels, divId: this.state.divId });
      this.raiseEvent();
      return panelKey;
    }

    /**Удаляет верхнюю панель из стека */
    removePanel(): string {
      if (this.panelsCount === 0) {
        return;
      }
      let panel = this.state.panels[this.state.panels.length - 1];
      let element: any = panel.element;

      if (!panel.controller || !panel.controller.canClose) {
        // Если не полноценная панель - просто удаляем
        this.state.panels.pop();
        this.setState({ panels: this.state.panels, divId: this.state.divId });
        this.raiseEvent();
        return panel.panelKey;
      }

      if (panel.controller.canClose && panel.controller.canClose(panel.panelKey)) {
        this.state.panels.pop();
        this.setState({ panels: this.state.panels, divId: this.state.divId });
        this.raiseEvent();
        return;
      }
      return panel.panelKey;
    }

    private _overlapStackPanelsChanged: event.Event<event.EventArgs>;
    public get overlapStackPanelsChanged(): event.Event<event.EventArgs> {
      return this._overlapStackPanelsChanged;
    }

    private raiseEvent() {
      this._overlapStackPanelsChanged.trigger(this, new event.EventArgs());
    }

    /**Признак полной инициализации компонента */
    private initialized: boolean = false;

    /**Компонент обновлент */
    componentDidUpdate(prevProps: IOverlapStackPanelsProps, prevState: State) {
      if (!this.initialized) {
        // Зарегистрирую содержимое
        this.addInitialChildrenPanel();
        if (this.props.initialized) {
          this.props.initialized(this);
        }
        this.initialized = true;
      }
    }

    /**Генерирует случайную строку вида abcd3333, подпадающую под условие ^[a-z]{2,5}\d{6}$ */
    static generateId(): string {
      let s = '';
      let iMax = 2 + Math.floor(Math.random() * 4);
      for (let i = 0; i < iMax; i++) {
        s += OverlapStackPanels.getStr('abcdefghijklmnopqrstuvwxyz');
      }
      for (let i = 0; i <= 5; i++) {
        s += OverlapStackPanels.getStr('0123456789');
      }
      return s;
    }

    /**Выдает случайный элемент из строки */
    static getStr(s: string): string {
      return s.charAt(Math.floor(Math.random() * s.length));
    }

    /**Проверяет реализуется ли объект методы интерфейса IOverlapPanelController */
    static IsOverlapPanelController(cls: any): boolean {
      // SIMA Потенциально, здесь может возникнуть проблема связанная с тем, что Object.getOwnPropertyDescriptor возвращает описание свойства
      // только текущего объекта, без учета наследования.
      // Поэтому, может потребоваться проверять свойства и методы через из поиск и в предках, получаемых методом Object.getPrototypeOf
      return cls.canClose !== 'undefined'
        && (cls.title !== 'undefined' || Object.getOwnPropertyDescriptor(cls, 'title') !== undefined)
        && (cls.overlapStackPanels !== 'undefined' || Object.getOwnPropertyDescriptor(cls, 'overlapStackPanels') !== undefined);
    }

    /**Добавление инициирующего содержимого (того, что было прописано в JSX'е, а не программно)*/
    addInitialChildrenPanel() {
      if (!this.props.children || this.state.panels.length > 0) {
        return;
      }
      React.Children.forEach(this.props.children, (child: JSX.Element) => {
        // this.state.panels.push(child);
        if (child.props && child.props.hasOwnProperty('overlapPanelController')) {
          this.addPanel(child, child.props.overlapPanelController);
        } else {
          this.addPanel(child);
        }

      });
    }

    /**Переназначает размеры для всех вложенных элементов (контейнеров для содержимого) */
    onContentResize(size: { height: number, width: number }) {
      this.forceUpdate();
    }

    render() {
      let parentElem = document.getElementById(this.state.divId);
      let className = css['overlap-panel'];
      let idx: number = 0;
      return (
        <AutoSizer onResize={this.onContentResize.bind(this)}>{(sz: ui.AutoSizerParams) => (
          <div data-overlapstackpanels id={this.state.divId} style={{ height: sz.height, width: sz.width }}>
            {this.state.panels.map((item: IOverlapPanelInfo, index: number, arr: IOverlapPanelInfo[]) => {
              if (index < arr.length - 1) {
                return (
                  <div key={item.panelKey} id={item.panelKey} className={className} style={{ height: parentElem.style.height, width: parentElem.style.width, position: 'absolute', visibility: 'collapse' }}>
                    {item.element}
                  </div>
                );
              }
              else {
                return (
                  <div key={item.panelKey} id={item.panelKey} className={className} style={{ height: parentElem.style.height, width: parentElem.style.width, position: 'absolute' }}>
                    {item.element}
                  </div>
                );
              }
            })}
          </div>
        )}
        </AutoSizer>
      );
    }
  }

}