// Это чтобы не материлась
declare function require(moduleName: string): any;

const store: any = require('core/Store').default;

type eventCallback = (...params: any[]) => any;

// =================================== IListener ==================================
export interface IListener {
  /**Register a function to be called when an event occurs */
  on(eventName: string, callback: eventCallback): void;

  /**Register a function to be called once when an event occurs. After being called the callback is unregistered */
  once(eventName: string, callback: eventCallback): void;

  /**Can unregister all callbacks from a listener if the `eventName` parameter is omitted, or all the callbacks for a `eventName` if the `callback` parameter is omitted. */
  off(eventName?: string, callback?: eventCallback): void;

  /**Trigger an event on the listener. All the extra parameters will be passed to the registered callbacks */
  trigger(eventName: string, ...params: any[]): void;
}

// ==================================== Store =====================================
export default class Store {
  /** Возвращает хранилище локального пользователя */
  static getP(): any {
    return store.getP();
  }

  /**Задает значение в локальном хранилище пользователя */
  static setP(keyOrObject: string | number | {}, value?: any): any {
    if (!value) {
      return store.setP(keyOrObject);
    }
    else {
      return store.setP(keyOrObject, value);
    }
  }

  static get(): any {
    return store.get();
  }

  static set(data: any) {
    store.set(data);
  }

  // Listener API

  /**Register a function to be called when an event occurs */
  static on(eventName: string, callback: eventCallback): void {
    store(eventName, callback);
  }

  /**Register a function to be called once when an event occurs. After being called the callback is unregistered */
  static once(eventName: string, callback: eventCallback): void {
    store.once(eventName, callback);
  }

  /**Can unregister all callbacks from a listener if the `eventName` parameter is omitted, or all the callbacks for a `eventName` if the `callback` parameter is omitted. */
  static off(eventName?: string, callback?: eventCallback): void {
    if (!eventName) {
      store.off(callback);
    }
    else if (!callback) {
      store.off(eventName);
    }
    else {
      store(eventName, callback);
    }
  }

  /**Trigger an event on the listener. All the extra parameters will be passed to the registered callbacks */
  static trigger(eventName: string, ...params: any[]): void {
      store.trigger(eventName, ...params);
  }

}
